package modManager;

import redguard.RtxDatabase;

import javax.swing.*;
import javax.swing.filechooser.FileNameExtensionFilter;
import java.awt.*;
import java.io.*;
import java.util.ArrayList;
import java.util.List;

public class SubtitleEditor {
    private static final String TITLE = "Redguard Subtitle Editor";

    // GUI fields
    private JFrame window;
    private List<JTextField> textFields;
    private JScrollPane scrollPane;
    private JTextField lastFoundText;
    private int foundIndex;
    private JTextField findTextField;

    // RTX Database
    private RtxDatabase rtxDatabase;

    public SubtitleEditor(RtxDatabase rtxDatabase) {
        this.rtxDatabase = rtxDatabase;
        createAndShowGUI();
    }

    private void createAndShowGUI() {
        window = new JFrame(TITLE);
        window.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        window.setMinimumSize(new Dimension(800, 600));
        textFields = new ArrayList<>();

        // Split GUI creation into sections
        createMenuBar();
        createTextFields();
        createBottomPanel();

        window.pack();
        window.setLocationRelativeTo(null);
        window.setVisible(true);
    }

    /**
     * Create the menu bar at the top with items for various commands.
     */
    private void createMenuBar() {
        JMenuBar menuBar = new JMenuBar();
        window.setJMenuBar(menuBar);

        // File menu
        JMenu fileMenu = new JMenu("File");
        fileMenu.setMnemonic('f');
        menuBar.add(fileMenu);
        // File menu items
        ModManagerUtils.createMenuItem(fileMenu, "Save RTX...", evt -> saveRtxFile());
        fileMenu.addSeparator();
        ModManagerUtils.createMenuItem(fileMenu, "Load Changes...", evt -> loadChangesFile());
        ModManagerUtils.createMenuItem(fileMenu, "Save Changes...", evt -> saveChangesFile());
        fileMenu.addSeparator();
        ModManagerUtils.createMenuItem(fileMenu, "Exit Subtitle Editor", evt -> window.dispose());
    }

    /**
     * Create the list of text fields that takes up the majority of the editor window
     */
    private void createTextFields() {
        SpringLayout layout = new SpringLayout();
        JPanel textFieldPanel = new JPanel(layout);
        textFieldPanel.setPreferredSize(new Dimension(0, 10 + rtxDatabase.getRtxEntries().size() * 29));

        for (int i = 0; i < rtxDatabase.getRtxEntries().size(); i++) {
            JLabel label = new JLabel(rtxDatabase.getRtxEntries().get(i).getLabel());
            if (i == 0) {
                layout.putConstraint(SpringLayout.NORTH, label, 10, SpringLayout.NORTH, textFieldPanel);
            } else {
                layout.putConstraint(SpringLayout.NORTH, label, 10, SpringLayout.SOUTH, textFields.get(i - 1));
            }
            layout.putConstraint(SpringLayout.WEST, label, 5, SpringLayout.WEST, textFieldPanel);
            textFieldPanel.add(label);

            JTextField textField = new JTextField();
            layout.putConstraint(SpringLayout.NORTH, textField, 0, SpringLayout.NORTH, label);
            layout.putConstraint(SpringLayout.WEST, textField, 35, SpringLayout.WEST, textFieldPanel);
            layout.putConstraint(SpringLayout.EAST, textField, -5, SpringLayout.EAST, textFieldPanel);
            textField.setText(rtxDatabase.getRtxEntries().get(i).getSubtitle());
            textField.setCaretPosition(0);
            textFieldPanel.add(textField);
            textFields.add(textField);
        }

        scrollPane = new JScrollPane(textFieldPanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        scrollPane.setPreferredSize(new Dimension(800, 560));
        scrollPane.getVerticalScrollBar().setUnitIncrement(4);
        scrollPane.getVerticalScrollBar().addAdjustmentListener(evt -> foundIndex = (scrollPane.getVerticalScrollBar().getValue() + 20) / 30);
        window.add(scrollPane, BorderLayout.CENTER);
    }

    /**
     * Create bottom panel that holds the find text field and load/save buttons
     */
    private void createBottomPanel() {
        JPanel bottomPanel = new JPanel();
        bottomPanel.setPreferredSize(new Dimension(0, 40));
        window.add(bottomPanel, BorderLayout.SOUTH);

        // Text field for finding label or subtitle matches
        JLabel findLabel = new JLabel("Find");
        bottomPanel.add(findLabel);
        findTextField = new JTextField(10);
        findTextField.addActionListener(evt -> findText());
        bottomPanel.add(findTextField);

        // Buttons
        ModManagerUtils.createButton(bottomPanel, "Reload", evt -> loadRtxFile());
        ModManagerUtils.createButton(bottomPanel, "Load from Mod", evt -> loadFromMod());
        ModManagerUtils.createButton(bottomPanel, "Save to Mod", evt -> saveToMod());
    }

    /**
     * Search for the next label or text field starting at foundIndex that matches the text in findTextField. If
     * the last text field is reached, continue from the top to make sure everything is searched.
     */
    private void findText() {
        String findText = findTextField.getText().toLowerCase();
        for (int i = 0; i < rtxDatabase.getRtxEntries().size(); i++) {
            int entry = (foundIndex + i) % rtxDatabase.getRtxEntries().size();
            JTextField textField = textFields.get(entry);
            boolean found = false;

            // Check label
            if (rtxDatabase.getRtxEntries().get(entry).getLabel().toLowerCase().contains(findText)) {
                found = true;
            } else {
                // Check text field
                int index = textField.getText().toLowerCase().indexOf(findText);
                if (index != -1) {
                    found = true;
                    // Select found text
                    textField.requestFocus();
                    textField.setCaretPosition(index);
                    textField.moveCaretPosition(index + findText.length());
                    findTextField.requestFocus();
                }
            }

            if (found) {
                // Deselect last found text
                if (lastFoundText != null) {
                    lastFoundText.setCaretPosition(0);
                    lastFoundText.moveCaretPosition(0);
                }
                lastFoundText = textField;

                scrollPane.getVerticalScrollBar().setValue(textFields.get(entry).getY());
                foundIndex = entry + 1;
                break;
            }
        }
    }

    private void loadRtxFile() {
        int confirmOption = JOptionPane.showConfirmDialog(window,
                "Are you sure you want to reload? All unsaved changes will be lost.", "Confirm reload",
                JOptionPane.YES_NO_OPTION);
        if (confirmOption == JOptionPane.YES_OPTION) {
            for (int i = 0; i < rtxDatabase.getRtxEntries().size(); i++) {
                textFields.get(i).setText(rtxDatabase.getRtxEntries().get(i).getSubtitle());
                textFields.get(i).setCaretPosition(0);
            }
        }
    }

    /**
     * Ask for the directory to save the RTX file, then use the public method
     */
    private void saveRtxFile() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Save RTX");
        chooser.setSelectedFile(new File("ENGLISH"));
        int option = chooser.showSaveDialog(window);
        if (option == JFileChooser.APPROVE_OPTION) {
            File file = ModManagerUtils.forceFileExtension(chooser.getSelectedFile(), ".RTX");
            if (ModManagerUtils.confirmReplace(window, file)) {
                RtxDatabase modifiedDatabase = new RtxDatabase(rtxDatabase);
                for (int i = 0; i < modifiedDatabase.getRtxEntries().size(); i++) {
                    modifiedDatabase.getRtxEntries().get(i).setSubtitle(textFields.get(i).getText());
                }
                try {
                    modifiedDatabase.writeFile(file);
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Ask for the directory to load from, then use the public method
     */
    private void loadChangesFile() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Load Changes");
        chooser.setFileFilter(new FileNameExtensionFilter("Normal text files (*.txt)", "txt"));
        chooser.setAcceptAllFileFilterUsed(false);
        int option = chooser.showOpenDialog(window);
        if (option == JFileChooser.APPROVE_OPTION) {
            loadChangesFile(chooser.getSelectedFile());
        }
    }

    public void loadChangesFile(File fileToLoad) {
        try (BufferedReader reader = new BufferedReader(new FileReader(fileToLoad))) {
            for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                String[] entry = line.split("\t", -1);
                int index = Integer.parseInt(entry[0]);
                textFields.get(index).setText(entry[1]);
                textFields.get(index).setCaretPosition(0);
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Ask for the directory to save to, then use the public method
     */
    private void saveChangesFile() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Save Changes");
        chooser.setFileFilter(new FileNameExtensionFilter("Normal text files (*.txt)", "txt"));
        chooser.setAcceptAllFileFilterUsed(false);
        chooser.setSelectedFile(new File("RTX Changes"));
        int option = chooser.showSaveDialog(window);
        if (option == JFileChooser.APPROVE_OPTION) {
            File file = ModManagerUtils.forceFileExtension(chooser.getSelectedFile(), ".txt");
            if (ModManagerUtils.confirmReplace(window, file)) {
                saveChangesFile(file);
            }
        }
    }

    private void saveChangesFile(File fileToSave) {
        try (BufferedWriter writer = new BufferedWriter(new FileWriter(fileToSave))) {
            for (int i = 0; i < rtxDatabase.getRtxEntries().size(); i++) {
                String originalText = rtxDatabase.getRtxEntries().get(i).getSubtitle();
                String text = textFields.get(i).getText();
                if (!text.equals(originalText)) {
                    writer.write(i + "\t" + text);
                    writer.newLine();
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void loadFromMod() {
        Mod mod = RedguardModManager.showModSelectionDialog(window, "Choose a mod to load subtitle changes.",
                "Load from Mod");
        if (mod != null) {
            File changesFile = RedguardModManager.getModPath(mod).resolve(RedguardModManager.RTX_CHANGES).toFile();
            if (changesFile.exists()) {
                loadChangesFile(changesFile);
            } else {
                JOptionPane.showMessageDialog(window, "Mod does not have an " + RedguardModManager.RTX_CHANGES
                        + " file.", "No changes file", JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }

    private void saveToMod() {
        Mod mod = RedguardModManager.showModSelectionDialog(window, "Choose a mod to save subtitle changes.",
                "Save to Mod");
        if (mod != null) {
            File changesFile = RedguardModManager.getModPath(mod).resolve(RedguardModManager.RTX_CHANGES).toFile();
            saveChangesFile(changesFile);
            if (changesFile.length() == 0) {
                changesFile.delete();
            }
        }
    }
}